"""
  This software is governed by the CeCILL-B license under French law and
  abiding by the rules of distribution of free software.  You can  use,
  modify and/ or redistribute the software under the terms of the CeCILL-B
  license as circulated by CEA, CNRS and INRIA at the following URL
  "http://www.cecill.info".
  As a counterpart to the access to the source code and  rights to copy,
  modify and redistribute granted by the license, users are provided only
  with a limited warranty  and the software's author,  the holder of the
  economic rights,  and the successive licensors  have only  limited
  liability.
  In this respect, the user's attention is drawn to the risks associated
  with loading,  using,  modifying and/or developing or reproducing the
  software by the user in light of its specific status of free software,
  that may mean  that it is complicated to manipulate,  and  that  also
  therefore means  that it is reserved for developers  and  experienced
  professionals having in-depth computer knowledge. Users are therefore
  encouraged to load and test the software's suitability as regards their
  requirements in conditions enabling the security of their systems and/or
  data to be ensured and,  more generally, to use and operate it in the
  same conditions as regards security.
  The fact that you are presently reading this means that you have had
  knowledge of the CeCILL-B license and that you accept its terms.
"""

import os
import h5py
import numpy as np
from operator import add

def store2hdf53D(filename, datas, labels, create=True, startloc=None, chunksz=256) :
    '''
    Store patches in format HDF5
    Adapted from 2D stor2hdf5 of Matcaffe (Caffe for matlab)
    Caffe supports data's form : N*C*H*W*D (numberOfBatches,channels,heigh,width,depth) 
    
    ----------
     datas, labels : 5D array, 
         formed N*C*H*W*D matrix of images should be normalized (e.g. to lie between 0 and 1) beforehand
     create :  True or False (default: true)
         specifies whether to create file newly or to append to previously created file, 
         useful to store information in batches when a dataset is too big to be held in memory  
     startloc : (point at which to start writing data). 
         By default, 
         if create=1 (create mode), startloc.data=[1, 1, 1, 1, 1], and startloc.lab=[1, 1, 1, 1, 1]; 
         if create=0 (append mode), startloc.data=[K+1, 1, 1, 1, 1], and startloc.lab = [K+1, 1, 1, 1, 1]; 
         where K is the current number of samples stored in the HDF
     chunksz : integer  (default: 256)
         (used only in create mode):
         specifies number of samples to be stored per chunk (see HDF5 documentation on chunking) 
         for creating HDF5 files with unbounded maximum size - TLDR; 
         higher chunk sizes allow faster read-write operations                 
                     
    '''
    # verify that format is right
    dat_dims = datas.shape
    lab_dims = labels.shape
    num_samples = dat_dims[0]
    if (num_samples != lab_dims[0]):
        raise AssertionError, 'Number of samples should be matched between data and labels'    
            
    # Check Create mode and Startloc    
    if create == True:
        if os.path.isfile(filename):
            print 'Warning: replacing existing file', filename, '\n'
            os.remove(filename)
           
        file = h5py.File(filename, "w")
        dset = file.create_dataset("data", shape=dat_dims , dtype='float32', maxshape=(None,)+tuple(dat_dims[1:]), chunks=(np.long(chunksz),)+tuple(dat_dims[1:]))
        lset = file.create_dataset("label", shape=lab_dims , dtype='float32', maxshape=(None,)+tuple(lab_dims[1:]), chunks=(np.long(chunksz),)+tuple(lab_dims[1:]))
        dset[...] = datas
        lset[...] = labels
        if  startloc == None:
            startloc = {'dat':0 , 'lab':0}
            startloc['dat']= (0,) + tuple(np.zeros(len(dat_dims)-1,dtype='int'))
            startloc['lab']= (0,) + tuple(np.zeros(len(lab_dims)-1,dtype='int'))           
    
    else:   # append mode
        if  startloc == None:
            file = h5py.File(filename, "r")
            prev_dat_sz = file[file.keys()[0]].shape
            prev_lab_sz = file[file.keys()[1]].shape
            if ( prev_dat_sz[1:] != dat_dims[1:] ):
                raise AssertionError, 'Data dimensions must match existing dimensions in dataset'
            if ( prev_lab_sz[1:] != lab_dims[1:] ):
                raise AssertionError, 'Label dimensions must match existing dimensions in dataset'    
            startloc = {'dat':0 , 'lab':0}
            startloc['dat']= (prev_dat_sz[0],) + tuple(np.zeros(len(dat_dims)-1,dtype='int'))
            startloc['lab']= (prev_lab_sz[0],) + tuple(np.zeros(len(lab_dims)-1,dtype='int'))
            
    # Writing data
    if (datas.size) or (labels.size):
        file = h5py.File(filename, "r+")
        dset = file['/data']
        lset = file['/label']
        dset.resize(map(add, dat_dims,startloc['dat']))
        lset.resize(map(add, lab_dims,startloc['lab'])) 
        dset[startloc['dat'][0]:startloc['dat'][0]+dat_dims[0],...] = datas
        lset[startloc['lab'][0]:startloc['lab'][0]+lab_dims[0],...] = labels      
    else:    
        assert 'store2hdf5 need datas'
    curr_dat_sz=file[file.keys()[0]].shape
    curr_lab_sz=file[file.keys()[1]].shape
    file.close()   
    return [curr_dat_sz,curr_lab_sz]

class ProcessingTrainingSet(object):  
    def __init__(self, TrainingFilesText, batchSize, InputName = 'data', LabelName = 'label'):  
        self.Datas = []
        self.Labels = []
        self.iterationPerEpoch = 0
        self.batchSize = batchSize
        
        # Get set of files
        with open(TrainingFilesText , "r") as TrainingFiles:
            string = TrainingFiles.read()
        self.FileSet = string.split()
        
        # Get location of each file in set and number of iteration each epoch
        for FileIndex in range(len(self.FileSet)):
            with h5py.File(self.FileSet[FileIndex],'r') as hf:
            # Reading all data
                if InputName in hf.keys():
                    datas = np.array(hf.get(InputName))
                    self.Datas.append(datas)
                else:
                    raise AssertionError, 'Key name ' + InputName + 'does not exist !' 
                
                if LabelName in hf.keys():
                    self.Labels.append(np.array(hf.get(LabelName)))
                else:
                    raise AssertionError, 'Key name ' + LabelName + ' does not exist !'
            self.iterationPerEpoch = self.iterationPerEpoch + datas.shape[0] / self.batchSize 
             
        # Tranfer to array
        self.Datas = np.concatenate(np.asarray(self.Datas,dtype=np.float32))
        self.Datas = self.Datas.reshape(-1,
                                        self.Datas.shape[-4],
                                        self.Datas.shape[-3],
                                        self.Datas.shape[-2],
                                        self.Datas.shape[-1])
        
        self.Labels = np.concatenate(np.asarray(self.Labels,dtype=np.float32))
        self.Labels = self.Labels.reshape(-1,
                                          self.Labels.shape[-4],
                                          self.Labels.shape[-3],
                                          self.Labels.shape[-2],
                                          self.Labels.shape[-1])



"""
  This software is governed by the CeCILL-B license under French law and
  abiding by the rules of distribution of free software.  You can  use,
  modify and/ or redistribute the software under the terms of the CeCILL-B
  license as circulated by CEA, CNRS and INRIA at the following URL
  "http://www.cecill.info".
  As a counterpart to the access to the source code and  rights to copy,
  modify and redistribute granted by the license, users are provided only
  with a limited warranty  and the software's author,  the holder of the
  economic rights,  and the successive licensors  have only  limited
  liability.
  In this respect, the user's attention is drawn to the risks associated
  with loading,  using,  modifying and/or developing or reproducing the
  software by the user in light of its specific status of free software,
  that may mean  that it is complicated to manipulate,  and  that  also
  therefore means  that it is reserved for developers  and  experienced
  professionals having in-depth computer knowledge. Users are therefore
  encouraged to load and test the software's suitability as regards their
  requirements in conditions enabling the security of their systems and/or
  data to be ensured and,  more generally, to use and operate it in the
  same conditions as regards security.
  The fact that you are presently reading this means that you have had
  knowledge of the CeCILL-B license and that you accept its terms.
"""
import os

s=os.path.abspath(__file__).split("/")

wd="/".join(s[0:(len(s)-1)])


os.chdir(wd)

import numpy as np
import SimpleITK as sitk
import scipy.ndimage
from ast import literal_eval as make_tuple

import sys
sys.path.insert(0, os.getcwd()+'/utils')
from utils3d import shave3D
from utils3d import pad3D
from SegSRGAN import SegSRGAN
import progressbar


class SegSRGAN_test(object):
        def __init__(self,weights,patch1,patch2,patch3):
            
            self.patch1 = patch1
            self.patch2 = patch2
            self.patch3 = patch3
            self.prediction = None
            self.SegSRGAN = SegSRGAN(ImageRow = patch1, 
                                     ImageColumn = patch2, 
                                     ImageDepth = patch3)
            self.GeneratorModel = self.SegSRGAN.generator_model_for_pred()
            self.GeneratorModel.load_weights(weights, by_name=True)
            self.generator = self.SegSRGAN.generator()
            
        def get_patch(self):
            
            return self.patch
            
        def test_by_patch(self,TestImage,step=1):  
            
            # Init temp
            Height,Width,Depth = np.shape(TestImage)
            
            
            TempHRImage = np.zeros_like(TestImage)
            TempSeg = np.zeros_like(TestImage)
            WeightedImage = np.zeros_like(TestImage)
            
            i=0
            bar = progressbar.ProgressBar(maxval=len(np.arange(0,Height-self.patch1+1,step))*len(np.arange(0,Width-self.patch2+1,step))*len(np.arange(0,Depth-self.patch3+1,step))).start()
    
            for idx in range(0,Height-self.patch1+1,step):
                for idy in range(0,Width-self.patch2+1,step):
                    for idz in range(0,Depth-self.patch3+1,step):  
                        
                        # Cropping image
                        TestPatch = TestImage[idx:idx+self.patch1,idy:idy+self.patch2,idz:idz+self.patch3] 
                        ImageTensor = TestPatch.reshape(1,1,self.patch1,self.patch2,self.patch3).astype(np.float32)
                        PredictPatch =  self.generator.predict(ImageTensor, batch_size=1)
                        
                        # Adding
                        TempHRImage[idx:idx+self.patch1,idy:idy+self.patch2,idz:idz+self.patch3] += PredictPatch[0,0,:,:,:]
                        TempSeg [idx:idx+self.patch1,idy:idy+self.patch2,idz:idz+self.patch3] += PredictPatch[0,1,:,:,:]
                        WeightedImage[idx:idx+self.patch1,idy:idy+self.patch2,idz:idz+self.patch3] += np.ones_like(PredictPatch[0,0,:,:,:])
                        
                        i+=1
                        
                        bar.update(i)
                        
                        
            # Weight sum of patches
            print('Done !')
            EstimatedHR = TempHRImage/WeightedImage
            EstimatedSegmentation = TempSeg/WeightedImage
            return (EstimatedHR,EstimatedSegmentation)
    



@girder_job(title='Segmentation haute resolution du cortex cerebral')
@app.task(bind=True)

def segmentation(self,input_file_path,step,spline_order,high_resolution,path_output_cortex,patch=None,):
    
    # TestFile = path de l'image en entree
    # high_resolution = tuple des resolutions (par axe)
        
     # Check resolution
    NewResolution = make_tuple(high_resolution)
    if np.isscalar(NewResolution):
        NewResolution = (NewResolution,NewResolution,NewResolution)
    else:
        if len(NewResolution)!=3:
            raise AssertionError('Not support this resolution !')
            
#    # Check border removing
#    border = make_tuple(args.border)
#    if np.isscalar(border):
#        border = (border,border,border)
#    else:
#        if len(border)!=3:
#            raise AssertionError, 'Not support this border !' 
        
    
    
          
    # Read low-resolution image
    TestNifti = sitk.ReadImage(input_file_path)
    TestImage = np.swapaxes(sitk.GetArrayFromImage(TestNifti),0,2).astype('float32')
    TestImageMinValue = float(np.min(TestImage))
    TestImageMaxValue = float(np.max(TestImage))
    TestImageNorm = TestImage/TestImageMaxValue 
    
    # Check scale factor type
    UpScale = tuple(itema/itemb for itema,itemb in zip(TestNifti.GetSpacing(),NewResolution)) 
    
    
    # spline interpolation 
    InterpolatedImage = scipy.ndimage.zoom(TestImageNorm, 
                                           zoom = UpScale,
                                           order = spline_order) 
    
    if patch is not None : 

        
        patch1 = patch2 = patch3 = int(patch)
        
        border=((InterpolatedImage.shape[0]-int(patch))%step,(InterpolatedImage.shape[1]-int(patch))%step,(InterpolatedImage.shape[2]-int(patch))%step)
    
        # Shave border
        ShavedInterpolatedImage = shave3D(InterpolatedImage, border) # remove border of the image
        
    else :
        border=(InterpolatedImage.shape[0]%4,InterpolatedImage.shape[1]%4,InterpolatedImage.shape[2]%4)
        ShavedInterpolatedImage = shave3D(InterpolatedImage, border) # remove border of the image
        Height,Width,Depth = np.shape(ShavedInterpolatedImage)
        patch1 = Height
        patch2 = Width
        patch3 = Depth
        
    #Weight path : 
    
    weight_path=os.getcwd()+"/weights/SegSRGAN"
    
    # Loading weights
    SegSRGAN_test_instance = SegSRGAN_test(weight_path,patch1,patch2,patch3)
    
    # GAN 
    print("Testing : ")
    EstimatedHRImage, EstimatedCortex  = SegSRGAN_test_instance.test_by_patch(ShavedInterpolatedImage,step=args.step) # parcours de l'image avec le patch
    
    # Padding
    #on fait l'operation de saving a l'envers
    PaddedEstimatedHRImage = pad3D(EstimatedHRImage,border)
    EstimatedCortex = pad3D(EstimatedCortex,border)
    
    # SR image 
    EstimatedHRImageInverseNorm = PaddedEstimatedHRImage*TestImageMaxValue
    EstimatedHRImageInverseNorm[EstimatedHRImageInverseNorm <= TestImageMinValue] = TestImageMinValue    # Clear negative value
    OutputImage = sitk.GetImageFromArray(np.swapaxes(EstimatedHRImageInverseNorm,0,2))
    OutputImage.SetSpacing(NewResolution)
    OutputImage.SetOrigin(TestNifti.GetOrigin())
    OutputImage.SetDirection(TestNifti.GetDirection())
    
    # Cortex segmentation
    OutputCortex = sitk.GetImageFromArray(np.swapaxes(EstimatedCortex,0,2))
    OutputCortex.SetSpacing(NewResolution)
    OutputCortex.SetOrigin(TestNifti.GetOrigin())
    OutputCortex.SetDirection(TestNifti.GetDirection())
    
    return "Segmentation Done"







import Function_for_application_test_python3
import pandas as pd
import glob
import os



###### Wrnings : 

# Il faut que la script soit appliquer a plus d'un fichier car path_pour_application doit etre une liste

###### Argument :

Descripif_poids="poids intiaux" # string qui apparaitra dans le dossier cree pour stocker les result : code avec + Descripif_poids

path_relatif_poids="/weights/SegSRGAN"

debut_relatif_path="/home/qdelannoy/Segmentation_neural_network/" #Path qui amene au fichier Base_pour_romeo

data=pd.read_csv(debut_relatif_path+"Base_pour_romeo/Meta_data_data_for_romeo.csv")

data=data[::-1] #renversement de l ordre

path_pour_application=debut_relatif_path+data["Path_relatif_pour_romeo"]


ensemble_pas=[]

#ensemble_pas.append([20,20])
#
#ensemble_pas.append([5,20]) #list des step qui vont etre appliquer pour la premiere taille de patch  
#
#ensemble_pas.append([10,40]) #multiplication de la taille du patch par 2, multiplication du pas par 2

ensemble_pas.append([40])

patchs=[128]


####### fonction : 

def createFolder(directory):
    try:
        if not os.path.exists(directory):
            os.makedirs(directory)
    except OSError:
        print ('Error: Creating directory. ' +  directory)


######## Execution : 

ensemble_pas=pd.DataFrame(ensemble_pas,index=patchs)

np_pas_per_patch=ensemble_pas.shape[1]

weights_path =  os.getcwd()+path_relatif_poids




for i in path_pour_application :
    
    for patch in patchs :

        if patch is None : 
            
            i_split=i.split("/")

            path_output = "/".join(i_split[:(len(i_split)-1)])+"/code avec "+Descripif_poids+"/patch "+str(patch)
            
            print(path_output)
               
            if  not os.path.exists(path_output):
                
                createFolder(path_output)
                    
                path_output_cortex = path_output + "/Cortex "+str(patch)+".nii.gz"
                    
                path_output_SR=path_output+"/SR " +str(patch)+".nii.gz"
                
                Function_for_application_test_python3.segmentation(input_file_path=i,
                                                                   step=20,
                                                                   NewResolution=(0.5,0.5,0.5),
                                                                   patch=patch,
                                                                   path_output_cortex=path_output_cortex,
                                                                   path_output_HR=path_output_SR,
                                                                   weights_path=weights_path
                                                                   )
            else :
                
                print("déja calculé")
        else :
    
            for step in ensemble_pas.loc[patch]:

                i_split=i.split("/")
    
                path_output = "/".join(i_split[:(len(i_split)-1)])+"/code avec "+Descripif_poids+"/patch "+str(patch)+" step "+str(step)+" inversion shave padd"
                    
                print(path_output)
                    
                if  not os.path.exists(path_output):
                            
                    createFolder(path_output)
                        
                    path_output_cortex = path_output + "/Cortex patch "+str(patch)+" step "+str(step)+".nii.gz"
                        
                    path_output_SR=path_output+"/SR.nii.gz"+str(patch)+" step "+str(step)+".nii.gz"
                    
                    Function_for_application_test_python3.segmentation(input_file_path=i,
                                                                       step=step,
                                                                       NewResolution=(0.5,0.5,0.5),
                                                                       patch=patch,
                                                                       path_output_cortex=path_output_cortex,
                                                                       path_output_HR=path_output_SR,
                                                                       weights_path=weights_path
                                                                       )
                else :
                        
                    print("déja calculé")
                




import os

import progressbar
import sys

import numpy as np
import SimpleITK as sitk
import scipy.ndimage
from ast import literal_eval as make_tuple

s = os.path.abspath(__file__).split("/")
wd = "/".join(s[0:(len(s) - 1)])
os.chdir(wd)
sys.path.insert(0, os.getcwd() + '/utils')

from utils3d import shave3D
from utils3d import pad3D
from SegSRGAN import SegSRGAN


class SegSRGAN_test(object):

    def __init__(self, weights, patch1, patch2, patch3, is_conditional, resolution=0):

        self.patch1 = patch1
        self.patch2 = patch2
        self.patch3 = patch3
        self.prediction = None
        self.SegSRGAN = SegSRGAN(ImageRow=patch1,
                                 ImageColumn=patch2,
                                 ImageDepth=patch3, is_conditional=is_conditional)
        self.GeneratorModel = self.SegSRGAN.generator_model_for_pred()
        self.GeneratorModel.load_weights(weights, by_name=True)
        self.generator = self.SegSRGAN.generator()
        self.is_conditional = is_conditional
        self.resolution = resolution
        self.ResTensor = np.expand_dims(np.expand_dims(np.ones([patch1, patch2, patch3]) * self.resolution, axis=0),
                                        axis=0)

    def get_patch(self):

        return self.patch

    def test_by_patch(self, TestImage, step=1, by_batch=False):

        # Init temp
        Height, Width, Depth = np.shape(TestImage)

        TempHRImage = np.zeros_like(TestImage)
        TempSeg = np.zeros_like(TestImage)
        WeightedImage = np.zeros_like(TestImage)

        # if is_conditional is set to True we predict on the image AND the resolution
        if self.is_conditional is True:
            if not by_batch:

                i = 0
                bar = progressbar.ProgressBar(maxval=len(np.arange(0, Height - self.patch1 + 1, step)) * len(
                    np.arange(0, Width - self.patch2 + 1, step)) * len(np.arange(0, Depth - self.patch3 + 1, step))).start()
                print('Patch=', self.patch1)
                print('Step=', step)
                for idx in range(0, Height - self.patch1 + 1, step):
                    for idy in range(0, Width - self.patch2 + 1, step):
                        for idz in range(0, Depth - self.patch3 + 1, step):
                            # Cropping image
                            TestPatch = TestImage[idx:idx + self.patch1, idy:idy + self.patch2, idz:idz + self.patch3]
                            ImageTensor = TestPatch.reshape(1, 1, self.patch1, self.patch2, self.patch3).astype(np.float32)
                            PredictPatch = self.generator.predict([ImageTensor, self.ResTensor], batch_size=1)

                            # Adding
                            TempHRImage[idx:idx + self.patch1, idy:idy + self.patch2,
                            idz:idz + self.patch3] += PredictPatch[0, 0, :, :, :]
                            TempSeg[idx:idx + self.patch1, idy:idy + self.patch2, idz:idz + self.patch3] += PredictPatch[0,
                                                                                                            1, :, :, :]
                            WeightedImage[idx:idx + self.patch1, idy:idy + self.patch2,
                            idz:idz + self.patch3] += np.ones_like(PredictPatch[0, 0, :, :, :])

                            i += 1

                            bar.update(i)
            else:

                Height = TestImage.shape[0]
                Width = TestImage.shape[1]
                Depth = TestImage.shape[2]

                patch1 = self.patch1
                patch2 = self.patch2
                patch3 = self.patch3

                patches = np.array([[TestImage[idx:idx + patch1, idy:idy + patch2, idz:idz + patch3]] for idx in
                                    range(0, Height - patch1 + 1, step) for idy in range(0, Width - patch2 + 1, step) for
                                    idz in range(0, Depth - patch3 + 1, step)])

                indice_patch = np.array([(idx, idy, idz) for idx in range(0, Height - patch1 + 1, step) for idy in
                                         range(0, Width - patch2 + 1, step) for idz in range(0, Depth - patch3 + 1, step)])

                pred = self.generator.predict(patches, batch_size=patches.shape[0])

                Weight = np.zeros_like(TestImage)
                TempHRImage = np.zeros(TestImage)
                TempSeg = np.zeros(TestImage)

                for i in range(indice_patch.shape[0]):
                    TempHRImage[indice_patch[i][0]:indice_patch[i][0] + patch1,
                    indice_patch[i][1]:indice_patch[i][1] + patch2, indice_patch[i][2]:indice_patch[i][2] + patch3] += pred[
                                                                                                                       i, 0,
                                                                                                                       :, :,
                                                                                                                       :]
                    TempSeg[indice_patch[i][0]:indice_patch[i][0] + patch1, indice_patch[i][1]:indice_patch[i][1] + patch2,
                    indice_patch[i][2]:indice_patch[i][2] + patch3] += pred[i, 1, :, :, :]
                    Weight[indice_patch[i][0]:indice_patch[i][0] + patch1, indice_patch[i][1]:indice_patch[i][1] + patch2,
                    indice_patch[i][2]:indice_patch[i][2] + patch3] + np.ones_like(
                        Weight[indice_patch[i][0]:indice_patch[i][0] + patch1,
                        indice_patch[i][1]:indice_patch[i][1] + patch2, indice_patch[i][2]:indice_patch[i][2] + patch3])
        else:
            if not by_batch:

                i = 0
                bar = progressbar.ProgressBar(maxval=len(np.arange(0, Height - self.patch1 + 1, step)) * len(
                    np.arange(0, Width - self.patch2 + 1, step)) * len(
                    np.arange(0, Depth - self.patch3 + 1, step))).start()
                print('Patch=', self.patch1)
                print('Step=', step)
                for idx in range(0, Height - self.patch1 + 1, step):
                    for idy in range(0, Width - self.patch2 + 1, step):
                        for idz in range(0, Depth - self.patch3 + 1, step):
                            # Cropping image
                            TestPatch = TestImage[idx:idx + self.patch1, idy:idy + self.patch2, idz:idz + self.patch3]
                            ImageTensor = TestPatch.reshape(1, 1, self.patch1, self.patch2, self.patch3).astype(
                                np.float32)
                            PredictPatch = self.generator.predict(ImageTensor, batch_size=1)

                            # Adding
                            TempHRImage[idx:idx + self.patch1, idy:idy + self.patch2,
                            idz:idz + self.patch3] += PredictPatch[0, 0, :, :, :]
                            TempSeg[idx:idx + self.patch1, idy:idy + self.patch2,
                            idz:idz + self.patch3] += PredictPatch[0,
                                                      1, :, :, :]
                            WeightedImage[idx:idx + self.patch1, idy:idy + self.patch2,
                            idz:idz + self.patch3] += np.ones_like(PredictPatch[0, 0, :, :, :])

                            i += 1

                            bar.update(i)
            else:

                Height = TestImage.shape[0]
                Width = TestImage.shape[1]
                Depth = TestImage.shape[2]

                patch1 = self.patch1
                patch2 = self.patch2
                patch3 = self.patch3

                patches = np.array([[TestImage[idx:idx + patch1, idy:idy + patch2, idz:idz + patch3]] for idx in
                                    range(0, Height - patch1 + 1, step) for idy in range(0, Width - patch2 + 1, step)
                                    for
                                    idz in range(0, Depth - patch3 + 1, step)])

                indice_patch = np.array([(idx, idy, idz) for idx in range(0, Height - patch1 + 1, step) for idy in
                                         range(0, Width - patch2 + 1, step) for idz in
                                         range(0, Depth - patch3 + 1, step)])

                pred = self.generator.predict(patches, batch_size=patches.shape[0])

                Weight = np.zeros_like(TestImage)
                TempHRImage = np.zeros(TestImage)
                TempSeg = np.zeros(TestImage)

                for i in range(indice_patch.shape[0]):
                    TempHRImage[indice_patch[i][0]:indice_patch[i][0] + patch1,
                    indice_patch[i][1]:indice_patch[i][1] + patch2,
                    indice_patch[i][2]:indice_patch[i][2] + patch3] += pred[
                                                                       i, 0,
                                                                       :, :,
                                                                       :]
                    TempSeg[indice_patch[i][0]:indice_patch[i][0] + patch1,
                    indice_patch[i][1]:indice_patch[i][1] + patch2,
                    indice_patch[i][2]:indice_patch[i][2] + patch3] += pred[i, 1, :, :, :]
                    Weight[indice_patch[i][0]:indice_patch[i][0] + patch1,
                    indice_patch[i][1]:indice_patch[i][1] + patch2,
                    indice_patch[i][2]:indice_patch[i][2] + patch3] + np.ones_like(
                        Weight[indice_patch[i][0]:indice_patch[i][0] + patch1,
                        indice_patch[i][1]:indice_patch[i][1] + patch2, indice_patch[i][2]:indice_patch[i][2] + patch3])
        # Weight sum of patches
        print('Done !')
        EstimatedHR = TempHRImage / WeightedImage
        EstimatedSegmentation = TempSeg / WeightedImage

        return (EstimatedHR, EstimatedSegmentation)


def segmentation(input_file_path, step, NewResolution, path_output_cortex, path_output_HR, weights_path, patch=None,
                 spline_order=3, by_batch=False, is_conditional=False):
    # TestFile = path de l'image en entree
    # high_resolution = tuple des resolutions (par axe)

    # Check resolution
    if np.isscalar(NewResolution):
        NewResolution = (NewResolution, NewResolution, NewResolution)
    else:
        if len(NewResolution) != 3:
            raise AssertionError('Not support this resolution !')

    # Read low-resolution image
    TestNifti = sitk.ReadImage(input_file_path)
    TestImage = np.swapaxes(sitk.GetArrayFromImage(TestNifti), 0, 2).astype('float32')
    TestImageMinValue = float(np.min(TestImage))
    TestImageMaxValue = float(np.max(TestImage))
    TestImageNorm = TestImage / TestImageMaxValue


    resolution = get_resolution(input_file_path)

    # Check scale factor type
    UpScale = tuple(itema / itemb for itema, itemb in zip(TestNifti.GetSpacing(), NewResolution))

    # spline interpolation 
    InterpolatedImage = scipy.ndimage.zoom(TestImageNorm,
                                           zoom=UpScale,
                                           order=spline_order)

    if patch is not None:

        print("patch given")

        patch1 = patch2 = patch3 = int(patch)

        border = (
        int((InterpolatedImage.shape[0] - int(patch)) % step), int((InterpolatedImage.shape[1] - int(patch)) % step),
        int((InterpolatedImage.shape[2] - int(patch)) % step))

        border_to_add = (step - border[0], step - border[1], step - border[2])

        # padd border
        PaddedInterpolatedImage = pad3D(InterpolatedImage, border_to_add)  # remove border of the image

    else:
        border = (
        int(InterpolatedImage.shape[0] % 4), int(InterpolatedImage.shape[1] % 4), int(InterpolatedImage.shape[2] % 4))
        border_to_add = (step - border[0], step - border[1], step - border[2])

        PaddedInterpolatedImage = pad3D(InterpolatedImage, border_to_add)  # remove border of the image

        Height, Width, Depth = np.shape(PaddedInterpolatedImage)
        patch1 = Height
        patch2 = Width
        patch3 = Depth

    # Loading weights
    SegSRGAN_test_instance = SegSRGAN_test(weights_path, patch1, patch2, patch3, is_conditional, resolution)

    # GAN 
    print("Testing : ")
    EstimatedHRImage, EstimatedCortex = SegSRGAN_test_instance.test_by_patch(PaddedInterpolatedImage, step=step,
                                                                             by_batch=by_batch)
    # parcours de l'image avec le patch

    # Padding
    # on fait l'operation de padding a l'envers
    PaddedEstimatedHRImage = shave3D(EstimatedHRImage, border_to_add)
    EstimatedCortex = shave3D(EstimatedCortex, border_to_add)

    # SR image 
    EstimatedHRImageInverseNorm = PaddedEstimatedHRImage * TestImageMaxValue
    EstimatedHRImageInverseNorm[
        EstimatedHRImageInverseNorm <= TestImageMinValue] = TestImageMinValue  # Clear negative value
    OutputImage = sitk.GetImageFromArray(np.swapaxes(EstimatedHRImageInverseNorm, 0, 2))
    OutputImage.SetSpacing(NewResolution)
    OutputImage.SetOrigin(TestNifti.GetOrigin())
    OutputImage.SetDirection(TestNifti.GetDirection())

    sitk.WriteImage(OutputImage, path_output_HR)

    # Cortex segmentation
    OutputCortex = sitk.GetImageFromArray(np.swapaxes(EstimatedCortex, 0, 2))
    OutputCortex.SetSpacing(NewResolution)
    OutputCortex.SetOrigin(TestNifti.GetOrigin())
    OutputCortex.SetDirection(TestNifti.GetDirection())

    sitk.WriteImage(OutputCortex, path_output_cortex)

    return "Segmentation Done"


def get_resolution(filename):
    reader = sitk.ImageFileReader()

    reader.SetFileName(filename)
    reader.LoadPrivateTagsOn();

    reader.ReadImageInformation();

    resolution = float(reader.GetMetaData('pixdim[3]'))

    return resolution



"""
  This software is governed by the CeCILL-B license under French law and
  abiding by the rules of distribution of free software.  You can  use,
  modify and/ or redistribute the software under the terms of the CeCILL-B
  license as circulated by CEA, CNRS and INRIA at the following URL
  "http://www.cecill.info".
  As a counterpart to the access to the source code and  rights to copy,
  modify and redistribute granted by the license, users are provided only
  with a limited warranty  and the software's author,  the holder of the
  economic rights,  and the successive licensors  have only  limited
  liability.
  In this respect, the user's attention is drawn to the risks associated
  with loading,  using,  modifying and/or developing or reproducing the
  software by the user in light of its specific status of free software,
  that may mean  that it is complicated to manipulate,  and  that  also
  therefore means  that it is reserved for developers  and  experienced
  professionals having in-depth computer knowledge. Users are therefore
  encouraged to load and test the software's suitability as regards their
  requirements in conditions enabling the security of their systems and/or
  data to be ensured and,  more generally, to use and operate it in the
  same conditions as regards security.
  The fact that you are presently reading this means that you have had
  knowledge of the CeCILL-B license and that you accept its terms.
"""

import numpy as np
from functools import partial

from keras.models import Model
from keras.layers import Input, LeakyReLU, Reshape
from keras.layers import Conv3D, Add, UpSampling3D, Activation, Concatenate
from keras.optimizers import Adam
from keras.initializers import lecun_normal
import keras.backend as K

gen_initializer = lecun_normal()

import sys

sys.path.insert(0, './utils')
from layers import wasserstein_loss, ReflectPadding3D, gradient_penalty_loss, InstanceNormalization3D, \
    activation_SegSRGAN, charbonnier_loss
from Adam_lr_mult import LR_Adam
from keras import losses
from keras.utils import multi_gpu_model
from tensorflow.python.client import device_lib


def get_available_gpus():
    local_device_protos = device_lib.list_local_devices()
    return [x.name for x in local_device_protos if x.device_type == 'GPU']


def resnet_blocks(input_res, kernel, name):
    in_res_1 = ReflectPadding3D(padding=1)(input_res)
    out_res_1 = Conv3D(kernel, 3, strides=1, kernel_initializer=gen_initializer,
                       use_bias=False,
                       name=name + '_conv_a',
                       data_format='channels_first')(in_res_1)
    out_res_1 = InstanceNormalization3D(name=name + '_isnorm_a')(out_res_1)
    out_res_1 = Activation('relu')(out_res_1)

    in_res_2 = ReflectPadding3D(padding=1)(out_res_1)
    out_res_2 = Conv3D(kernel, 3, strides=1, kernel_initializer=gen_initializer,
                       use_bias=False,
                       name=name + '_conv_b',
                       data_format='channels_first')(in_res_2)
    out_res_2 = InstanceNormalization3D(name=name + '_isnorm_b')(out_res_2)

    out_res = Add()([out_res_2, input_res])
    return out_res


class SegSRGAN(object):
    def __init__(self, ImageRow=64, ImageColumn=64, ImageDepth=64,
                 FirstDiscriminatorKernel=32, FirstGeneratorKernel=16,
                 lamb_rec=1, lamb_adv=0.001, lamb_gp=10,
                 lr_DisModel=0.0001, lr_GenModel=0.0001, u_net_gen=False, multi_gpu=True,
                 is_conditional=False,
                 is_residual=True):
        self.ImageRow = ImageRow
        self.ImageColumn = ImageColumn
        self.ImageDepth = ImageDepth
        self.D = None  # discriminator
        self.G = None  # generator
        self.DisModel = None  # discriminator model
        self.DisModel_multi_gpu = None
        self.GenModel = None  # generator model
        self.GenModel_multi_gpu = None
        self.DiscriminatorKernel = FirstDiscriminatorKernel  # profondeur des carac extraite pour le gen
        self.GeneratorKernel = FirstGeneratorKernel  # profondeur des carac extraites pour le discri
        self.lamb_adv = lamb_adv
        self.lamb_rec = lamb_rec
        self.lamb_gp = lamb_gp
        self.lr_DisModel = lr_DisModel
        self.lr_GenModel = lr_GenModel
        self.u_net_gen = u_net_gen
        self.multi_gpu = multi_gpu
        self.is_conditional = is_conditional
        self.is_residual = is_residual

    def discriminator_block(self, name):
        """Creates a discriminator model that takes an image as input and outputs a single value, representing whether
        the input is real or generated. Unlike normal GANs, the output is not sigmoid and does not represent a
        probability!
        Instead, the output should be as large and negative as possible for generated inputs and as large and positive
        as possible for real inputs.
        Note that the improved WGAN paper suggests that BatchNormalization should not be used in the discriminator."""

        # In:
        inputs = Input(shape=(2, self.ImageRow, self.ImageColumn, self.ImageDepth), name='dis_input')

        # Input 64
        disnet = Conv3D(self.DiscriminatorKernel * 1, 4, strides=2,
                        padding='same',
                        kernel_initializer='he_normal',
                        data_format='channels_first',
                        name=name + '_conv_dis_1')(inputs)
        disnet = LeakyReLU(0.01)(disnet)

        # Hidden 1 : 32
        disnet = Conv3D(self.DiscriminatorKernel * 2, 4, strides=2,
                        padding='same',
                        kernel_initializer='he_normal',
                        data_format='channels_first',
                        name=name + '_conv_dis_2')(disnet)
        disnet = LeakyReLU(0.01)(disnet)

        # Hidden 2 : 16
        disnet = Conv3D(self.DiscriminatorKernel * 4, 4, strides=2,
                        padding='same',
                        kernel_initializer='he_normal',
                        data_format='channels_first',
                        name=name + '_conv_dis_3')(disnet)
        disnet = LeakyReLU(0.01)(disnet)

        # Hidden 3 : 8
        disnet = Conv3D(self.DiscriminatorKernel * 8, 4, strides=2,
                        padding='same',
                        kernel_initializer='he_normal',
                        data_format='channels_first',
                        name=name + '_conv_dis_4')(disnet)
        disnet = LeakyReLU(0.01)(disnet)

        # Hidden 4 : 4
        disnet = Conv3D(self.DiscriminatorKernel * 16, 4, strides=2,
                        padding='same',
                        kernel_initializer='he_normal',
                        data_format='channels_first',
                        name=name + '_conv_dis_5')(disnet)
        disnet = LeakyReLU(0.01)(disnet)

        # Decision : 2
        decision = Conv3D(1, 2, strides=1,
                          use_bias=False,
                          kernel_initializer='he_normal',
                          data_format='channels_first',
                          name='dis_decision')(disnet)

        decision = Reshape((1,))(decision)

        model = Model(inputs=[inputs], outputs=[decision], name=name)
        return model

    def generator_block(self, name):  # generateur meme dim en entree et sortie si multiple de 4
        #
        inputs = Input(shape=(1, self.ImageRow, self.ImageColumn, self.ImageDepth))

        # Representation
        gennet = ReflectPadding3D(padding=3)(inputs)
        gennet = Conv3D(self.GeneratorKernel, 7, strides=1, kernel_initializer=gen_initializer,
                        use_bias=False,
                        name=name + '_gen_conv1',
                        data_format='channels_first')(gennet)
        gennet = InstanceNormalization3D(name=name + '_gen_isnorm_conv1')(gennet)
        gennet = Activation('relu')(gennet)

        # Downsampling 1
        gennet = ReflectPadding3D(padding=1)(gennet)
        gennet = Conv3D(self.GeneratorKernel * 2, 3, strides=2, kernel_initializer=gen_initializer,
                        use_bias=False,
                        name=name + '_gen_conv2',
                        data_format='channels_first')(gennet)
        gennet = InstanceNormalization3D(name=name + '_gen_isnorm_conv2')(gennet)
        gennet = Activation('relu')(gennet)

        # Downsampling 2
        gennet = ReflectPadding3D(padding=1)(gennet)
        gennet = Conv3D(self.GeneratorKernel * 4, 3, strides=2, kernel_initializer=gen_initializer,
                        use_bias=False,
                        name=name + '_gen_conv3',
                        data_format='channels_first')(gennet)
        gennet = InstanceNormalization3D(name=name + '_gen_isnorm_conv3')(gennet)
        gennet = Activation('relu')(gennet)

        # Resnet blocks : 6, 8*4 = 32
        gennet = resnet_blocks(gennet, self.GeneratorKernel * 4, name=name + '_gen_block1')
        gennet = resnet_blocks(gennet, self.GeneratorKernel * 4, name=name + '_gen_block2')
        gennet = resnet_blocks(gennet, self.GeneratorKernel * 4, name=name + '_gen_block3')
        gennet = resnet_blocks(gennet, self.GeneratorKernel * 4, name=name + '_gen_block4')
        gennet = resnet_blocks(gennet, self.GeneratorKernel * 4, name=name + '_gen_block5')
        gennet = resnet_blocks(gennet, self.GeneratorKernel * 4, name=name + '_gen_block6')

        # Upsampling 1
        gennet = UpSampling3D(size=(2, 2, 2),
                              data_format='channels_first')(gennet)
        gennet = ReflectPadding3D(padding=1)(gennet)
        gennet = Conv3D(self.GeneratorKernel * 2, 3, strides=1, kernel_initializer=gen_initializer,
                        use_bias=False,
                        name=name + '_gen_deconv1',
                        data_format='channels_first')(gennet)
        gennet = InstanceNormalization3D(name=name + '_gen_isnorm_deconv1')(gennet)
        gennet = Activation('relu')(gennet)

        # Upsampling 2
        gennet = UpSampling3D(size=(2, 2, 2),
                              data_format='channels_first')(gennet)
        gennet = ReflectPadding3D(padding=1)(gennet)
        gennet = Conv3D(self.GeneratorKernel, 3, strides=1, kernel_initializer=gen_initializer,
                        use_bias=False,
                        name=name + '_gen_deconv2',
                        data_format='channels_first')(gennet)
        gennet = InstanceNormalization3D(name=name + '_gen_isnorm_deconv2')(gennet)
        gennet = Activation('relu')(gennet)

        # Reconstruction
        gennet = ReflectPadding3D(padding=3)(gennet)
        gennet = Conv3D(2, 7, strides=1, kernel_initializer=gen_initializer,
                        use_bias=False,
                        name=name + '_gen_1conv',
                        data_format='channels_first')(gennet)

        predictions = gennet
        predictions = activation_SegSRGAN(is_residual=self.is_residual)(
            [predictions, inputs])  # sigmoid proba + add input and pred SR

        model = Model(inputs=inputs, outputs=predictions, name=name)
        return model

    def discriminator_block_conditionnal(self, name):
        """Creates a discriminator model that takes an image as input and outputs a single value, representing whether
        the input is real or generated. Unlike normal GANs, the output is not sigmoid and does not represent a
        probability!
        Instead, the output should be as large and negative as possible for generated inputs and as large and positive
        as possible for real inputs.
        Note that the improved WGAN paper suggests that BatchNormalization should not be used in the discriminator."""

        # In:
        im = Input(shape=(2, self.ImageRow, self.ImageColumn, self.ImageDepth), name='dis_input')

        res = Input(shape=(1, self.ImageRow, self.ImageColumn, self.ImageDepth), name='dis_input_res')

        inputs = Concatenate(axis=-4)([im, res])

        # Input 64
        disnet = Conv3D(self.DiscriminatorKernel * 1, 4, strides=2,
                        padding='same',
                        kernel_initializer='he_normal',
                        data_format='channels_first',
                        name=name + '_conv_dis_1')(inputs)
        disnet = LeakyReLU(0.01)(disnet)

        # Hidden 1 : 32
        disnet = Conv3D(self.DiscriminatorKernel * 2, 4, strides=2,
                        padding='same',
                        kernel_initializer='he_normal',
                        data_format='channels_first',
                        name=name + '_conv_dis_2')(disnet)
        disnet = LeakyReLU(0.01)(disnet)

        # Hidden 2 : 16
        disnet = Conv3D(self.DiscriminatorKernel * 4, 4, strides=2,
                        padding='same',
                        kernel_initializer='he_normal',
                        data_format='channels_first',
                        name=name + '_conv_dis_3')(disnet)
        disnet = LeakyReLU(0.01)(disnet)

        # Hidden 3 : 8
        disnet = Conv3D(self.DiscriminatorKernel * 8, 4, strides=2,
                        padding='same',
                        kernel_initializer='he_normal',
                        data_format='channels_first',
                        name=name + '_conv_dis_4')(disnet)
        disnet = LeakyReLU(0.01)(disnet)

        # Hidden 4 : 4
        disnet = Conv3D(self.DiscriminatorKernel * 16, 4, strides=2,
                        padding='same',
                        kernel_initializer='he_normal',
                        data_format='channels_first',
                        name=name + '_conv_dis_5')(disnet)
        disnet = LeakyReLU(0.01)(disnet)

        # Decision : 2
        decision = Conv3D(1, 2, strides=1,
                          use_bias=False,
                          kernel_initializer='he_normal',
                          data_format='channels_first',
                          name='dis_decision')(disnet)

        decision = Reshape((1,))(decision)

        model = Model(inputs=[im, res], outputs=[decision], name=name)

        return model

    def generator_block_conditionnal(self, name):  # generateur meme dim en entree et sortie si multiple de 4
        #
        im = Input(shape=(1, self.ImageRow, self.ImageColumn, self.ImageDepth), name='dis_input')

        res = Input(shape=(1, self.ImageRow, self.ImageColumn, self.ImageDepth), name='dis_input_res')

        inputs = Concatenate(axis=-4)([im, res])

        # Representation
        gennet = ReflectPadding3D(padding=3)(inputs)
        gennet = Conv3D(self.GeneratorKernel, 7, strides=1, kernel_initializer=gen_initializer,
                        use_bias=False,
                        name=name + '_gen_conv1',
                        data_format='channels_first')(gennet)
        gennet = InstanceNormalization3D(name=name + '_gen_isnorm_conv1')(gennet)
        gennet = Activation('relu')(gennet)

        # Downsampling 1
        gennet = ReflectPadding3D(padding=1)(gennet)
        gennet = Conv3D(self.GeneratorKernel * 2, 3, strides=2, kernel_initializer=gen_initializer,
                        use_bias=False,
                        name=name + '_gen_conv2',
                        data_format='channels_first')(gennet)
        gennet = InstanceNormalization3D(name=name + '_gen_isnorm_conv2')(gennet)
        gennet = Activation('relu')(gennet)

        # Downsampling 2
        gennet = ReflectPadding3D(padding=1)(gennet)
        gennet = Conv3D(self.GeneratorKernel * 4, 3, strides=2, kernel_initializer=gen_initializer,
                        use_bias=False,
                        name=name + '_gen_conv3',
                        data_format='channels_first')(gennet)
        gennet = InstanceNormalization3D(name=name + '_gen_isnorm_conv3')(gennet)
        gennet = Activation('relu')(gennet)

        # Resnet blocks : 6, 8*4 = 32
        gennet = resnet_blocks(gennet, self.GeneratorKernel * 4, name=name + '_gen_block1')
        gennet = resnet_blocks(gennet, self.GeneratorKernel * 4, name=name + '_gen_block2')
        gennet = resnet_blocks(gennet, self.GeneratorKernel * 4, name=name + '_gen_block3')
        gennet = resnet_blocks(gennet, self.GeneratorKernel * 4, name=name + '_gen_block4')
        gennet = resnet_blocks(gennet, self.GeneratorKernel * 4, name=name + '_gen_block5')
        gennet = resnet_blocks(gennet, self.GeneratorKernel * 4, name=name + '_gen_block6')

        # Upsampling 1
        gennet = UpSampling3D(size=(2, 2, 2),
                              data_format='channels_first')(gennet)
        gennet = ReflectPadding3D(padding=1)(gennet)
        gennet = Conv3D(self.GeneratorKernel * 2, 3, strides=1, kernel_initializer=gen_initializer,
                        use_bias=False,
                        name=name + '_gen_deconv1',
                        data_format='channels_first')(gennet)
        gennet = InstanceNormalization3D(name=name + '_gen_isnorm_deconv1')(gennet)
        gennet = Activation('relu')(gennet)

        # Upsampling 2
        gennet = UpSampling3D(size=(2, 2, 2),
                              data_format='channels_first')(gennet)
        gennet = ReflectPadding3D(padding=1)(gennet)
        gennet = Conv3D(self.GeneratorKernel, 3, strides=1, kernel_initializer=gen_initializer,
                        use_bias=False,
                        name=name + '_gen_deconv2',
                        data_format='channels_first')(gennet)
        gennet = InstanceNormalization3D(name=name + '_gen_isnorm_deconv2')(gennet)
        gennet = Activation('relu')(gennet)

        # Reconstruction
        gennet = ReflectPadding3D(padding=3)(gennet)
        gennet = Conv3D(2, 7, strides=1, kernel_initializer=gen_initializer,
                        use_bias=False,
                        name=name + '_gen_1conv',
                        data_format='channels_first')(gennet)

        predictions = gennet
        predictions = activation_SegSRGAN(is_residual=self.is_residual)(
            [predictions, im])  # sigmoid proba + add input and pred SR

        model = Model(inputs=[im, res], outputs=predictions, name=name)

        return model

    def generator_block_u_net(self, name):  # generateur meme dim en entree et sortie si multiple de 4
        #
        inputs = Input(shape=(1, self.ImageRow, self.ImageColumn, self.ImageDepth))

        # Representation
        gennet = ReflectPadding3D(padding=3)(inputs)
        gennet = Conv3D(self.GeneratorKernel, 7, strides=1, kernel_initializer=gen_initializer,
                        use_bias=False,
                        name=name + '_gen_conv1',
                        data_format='channels_first')(gennet)
        gennet = InstanceNormalization3D(name=name + '_gen_isnorm_conv1')(gennet)
        gennet = Activation('relu')(gennet)

        # resblock :
        gennet = resnet_blocks(gennet, self.GeneratorKernel, name=name + '_gen_block')

        # Downsampling 1
        gennet_down_1 = ReflectPadding3D(padding=1)(gennet)
        gennet_down_1 = Conv3D(self.GeneratorKernel * 2, 3, strides=2, kernel_initializer=gen_initializer,
                               use_bias=False,
                               name=name + '_gen_conv2',
                               data_format='channels_first')(gennet_down_1)
        gennet_down_1 = InstanceNormalization3D(name=name + '_gen_isnorm_conv2')(gennet_down_1)
        gennet_down_1 = Activation('relu')(gennet_down_1)

        # resblock 1 :
        gennet_down_1 = resnet_blocks(gennet_down_1, self.GeneratorKernel * 2, name=name + '_gen_block1')

        # Downsampling 2
        gennet_down_2 = ReflectPadding3D(padding=1)(gennet_down_1)
        gennet_down_2 = Conv3D(self.GeneratorKernel * 4, 3, strides=2, kernel_initializer=gen_initializer,
                               use_bias=False,
                               name=name + '_gen_conv3',
                               data_format='channels_first')(gennet_down_2)
        gennet_down_2 = InstanceNormalization3D(name=name + '_gen_isnorm_conv3')(gennet_down_2)
        gennet_down_2 = Activation('relu')(gennet_down_2)

        # resblock 2
        gennet_down_2 = resnet_blocks(gennet_down_2, self.GeneratorKernel * 4, name=name + '_gen_block2')

        # Upsampling X2 down_2 : 

        gennet_up_1 = UpSampling3D(size=(2, 2, 2),
                                   data_format='channels_first')(gennet_down_2)
        gennet_up_1 = ReflectPadding3D(padding=1)(gennet_up_1)
        gennet_up_1 = Conv3D(self.GeneratorKernel * 2, 3, strides=1, kernel_initializer=gen_initializer,
                             use_bias=False,
                             name=name + '_gen_deconv1',
                             data_format='channels_first')(gennet_up_1)
        gennet_up_1 = InstanceNormalization3D(name=name + '_gen_isnorm_deconv1')(gennet_up_1)
        gennet_up_1 = Activation('relu')(gennet_up_1)

        #        del gennet_down_2

        # Concatenante gennet_up_1 with gennet_down_1 
        gennet_concate_1 = Concatenate(axis=-4)([gennet_up_1, gennet_down_1])

        #        del gennet_up_1
        #        del gennet_down_1

        # Upsampling 2
        gennet_up_2 = UpSampling3D(size=(2, 2, 2),
                                   data_format='channels_first')(gennet_concate_1)
        gennet_up_2 = ReflectPadding3D(padding=1)(gennet_up_2)
        gennet_up_2 = Conv3D(self.GeneratorKernel, 3, strides=1, kernel_initializer=gen_initializer,
                             use_bias=False,
                             name=name + '_gen_deconv2',
                             data_format='channels_first')(gennet_up_2)
        gennet_up_2 = InstanceNormalization3D(name=name + '_gen_isnorm_deconv2')(gennet_up_2)
        gennet_up_2 = Activation('relu')(gennet_up_2)

        # Concatenante gennet_up_2 with gennet_down_1
        gennet_concate_2 = Concatenate(axis=-4)([gennet_up_2, gennet])

        #        del gennet_concate_1
        #        del gennet_up_2

        # Reconstruction
        gennet_concate_2 = ReflectPadding3D(padding=3)(gennet_concate_2)
        gennet_concate_2 = Conv3D(2, 7, strides=1, kernel_initializer=gen_initializer,
                                  use_bias=False,
                                  name=name + '_gen_1conv',
                                  data_format='channels_first')(gennet_concate_2)

        predictions = gennet_concate_2
        predictions = activation_SegSRGAN(is_residual=self.is_residual)([predictions, inputs])

        model = Model(inputs=inputs, outputs=predictions, name=name)

        return model

    def generator(self):
        if self.G:
            return self.G
        if self.u_net_gen:
            self.G = self.generator_block_u_net('G')
        elif self.is_conditional:
            self.G = self.generator_block_conditionnal('G_cond')
        else:
            self.G = self.generator_block('G')

        return self.G

    def discriminator(self):
        if self.D:
            return self.D
        if self.is_conditional:
            self.D = self.discriminator_block_conditionnal('DX_cond')
        else:
            self.D = self.discriminator_block('DX')
        return self.D

    def generator_multi_gpu(self):

        num_gpu = len(get_available_gpus())

        if self.multi_gpu and (num_gpu > 1):

            gen_multi_gpu = multi_gpu_model(self.generator(), gpus=num_gpu, cpu_merge=False)
            print("Generator duplicated on : " + str(num_gpu) + " GPUS")
        else:
            gen_multi_gpu = self.generator()

        return gen_multi_gpu

    def discri_multi_gpu(self):

        num_gpu = len(get_available_gpus())

        if self.multi_gpu and (num_gpu > 1):

            discri_multi_gpu = multi_gpu_model(self.discriminator(), gpus=num_gpu, cpu_merge=False)
            print("Generator duplicated on : " + str(num_gpu) + " GPUS")
        else:
            discri_multi_gpu = self.discriminator()

        return discri_multi_gpu

    def generator_model(self):
        if self.GenModel:
            return self.GenModel

        print("We freeze the weights of Discriminator by setting their learning rate as 0 when updating Generator !")
        # We freeze the weights of Discriminator by setting their learning rate as 0 when updating Generator !
        #        AllParameters = 63
        #        GeneratorParameters = 52
        AllParameters = len(self.generator().get_weights()) + len(self.discriminator().get_weights())
        GeneratorParameters = len(self.generator().get_weights())
        multipliers = np.ones(AllParameters)
        for idx in range(GeneratorParameters, AllParameters):
            multipliers[idx] = 0.0

        input_im = Input(shape=(1, self.ImageRow, self.ImageColumn, self.ImageDepth), name='input_im_gen')

        if self.is_conditional:

            input_res = Input(shape=(1, self.ImageRow, self.ImageColumn, self.ImageDepth), name='input_res_gen')
            Gx_gen = self.generator()([input_im, input_res])  # Fake X
            fool_decision = self.discriminator()([Gx_gen, input_res])  # Fooling D
            # Model
            self.GenModel = Model([input_im, input_res], [fool_decision, Gx_gen])

        else:

            Gx_gen = self.generator()(input_im)  # Fake X
            fool_decision = self.discriminator()(Gx_gen)  # Fooling D
            # Model
            self.GenModel = Model(input_im, [fool_decision, Gx_gen])

        # print archi :

        self.generator().summary(line_length=150)

        num_gpu = len(get_available_gpus())

        if (self.multi_gpu) & (num_gpu > 1):

            self.GenModel_multi_gpu = multi_gpu_model(self.GenModel, gpus=num_gpu, cpu_merge=False)
            print("Generator Model duplicated on : " + str(num_gpu) + " GPUS")
        else:
            self.GenModel_multi_gpu = self.GenModel
            print("Generator Model apply on CPU or single GPU")

            # self.GenModel = multi_gpu_model(self.GenModel, gpus=num_gpu)
        self.GenModel_multi_gpu.compile(LR_Adam(lr=self.lr_GenModel, beta_1=0.5, beta_2=0.999, multipliers=multipliers),
                                        loss=[wasserstein_loss, charbonnier_loss],
                                        loss_weights=[self.lamb_adv, self.lamb_rec])

        return self.GenModel, self.GenModel_multi_gpu

    def generator_model_for_pred(self):
        if self.GenModel:
            return self.GenModel

        print("We freeze the weights of Discriminator by setting their learning rate as 0 when updating Generator !")
        # We freeze the weights of Discriminator by setting their learning rate as 0 when updating Generator !
        AllParameters = 63
        GeneratorParameters = 52
        multipliers = np.ones(AllParameters)
        for idx in range(GeneratorParameters, AllParameters):
            multipliers[idx] = 0.0

        # Input
        input_gen = Input(shape=(1, self.ImageRow, self.ImageColumn, self.ImageDepth), name='input_gen')

        if self.is_conditional:

            input_res = Input(shape=(1, self.ImageRow, self.ImageColumn, self.ImageDepth), name='input_res_gen')
            Gx_gen = self.generator()([input_gen, input_res])  # Fake X
            # Model
            self.GenModel = Model([input_gen, input_res], [Gx_gen])

        else:

            Gx_gen = self.generator()(input_gen)  # Fake X
            # Model
            self.GenModel = Model(input_gen, [Gx_gen])

        # ajout d'un loss quelconque car nous en avons pas besoin pour l'application d'un model mais il ne faut pas que
        # celui-ci soit definit
        # en utilisant de discri
        self.GenModel.compile(LR_Adam(lr=self.lr_GenModel, beta_1=0.5, beta_2=0.999, multipliers=multipliers),
                              losses.mean_squared_error)

        return self.GenModel

    def discriminator_model(self):
        if self.DisModel:
            return self.DisModel

            # Input
        real_dis = Input(shape=(2, self.ImageRow, self.ImageColumn, self.ImageDepth), name='real_dis')
        fake_dis = Input(shape=(2, self.ImageRow, self.ImageColumn, self.ImageDepth), name='fake_dis')
        interp_dis = Input(shape=(2, self.ImageRow, self.ImageColumn, self.ImageDepth), name='interp_dis')

        if self.is_conditional:

            res = Input(shape=(1, self.ImageRow, self.ImageColumn, self.ImageDepth))

            # Discriminator
            real_decision = self.discriminator()([real_dis, res])  # Real X
            fake_decision = self.discriminator()([fake_dis, res])  # Fake X
            interp_decision = self.discriminator()([interp_dis, res])  # interpolation X

        else:

            # Discriminator
            real_decision = self.discriminator()(real_dis)  # Real X
            fake_decision = self.discriminator()(fake_dis)  # Fake X
            interp_decision = self.discriminator()(interp_dis)  # interpolation X

        # GP loss
        partial_gp_loss = partial(gradient_penalty_loss,
                                  averaged_samples=interp_dis,
                                  gradient_penalty_weight=self.lamb_gp)
        partial_gp_loss.__name__ = 'gradient_penalty'  # Functions need names or Keras will throw an error*      

        num_gpu = len(get_available_gpus())

        print("number of gpus : " + str(num_gpu))

        if self.is_conditional:

            # Model
            self.DisModel = Model([real_dis, fake_dis, interp_dis, res],
                                  [real_decision, fake_decision, interp_decision])
        else:
            # Model
            self.DisModel = Model([real_dis, fake_dis, interp_dis], [real_decision, fake_decision, interp_decision])

        if (self.multi_gpu) & (num_gpu > 1):

            self.DisModel_multi_gpu = multi_gpu_model(self.DisModel, gpus=num_gpu, cpu_merge=False)
            print("Discriminator Model duplicated on : " + str(num_gpu) + " GPUS")
        else:
            self.DisModel_multi_gpu = self.DisModel
            print("Discriminator Model apply on CPU or single GPU")

        # self.DisModel = multi_gpu_model(self.DisModel, gpus=num_gpu)
        self.DisModel_multi_gpu.compile(Adam(lr=self.lr_DisModel, beta_1=0.5, beta_2=0.999),
                                        loss=[wasserstein_loss, wasserstein_loss, partial_gp_loss],
                                        loss_weights=[1, 1, self.lamb_gp])
        # multi gpu training ne change rien au temps d'exectution sur romeo. meme en changeant l'argument cpu_merge=False.

        return self.DisModel, self.DisModel_multi_gpu



"""
  This software is governed by the CeCILL-B license under French law and
  abiding by the rules of distribution of free software.  You can  use,
  modify and/ or redistribute the software under the terms of the CeCILL-B
  license as circulated by CEA, CNRS and INRIA at the following URL
  "http://www.cecill.info".
  As a counterpart to the access to the source code and  rights to copy,
  modify and redistribute granted by the license, users are provided only
  with a limited warranty  and the software's author,  the holder of the
  economic rights,  and the successive licensors  have only  limited
  liability.
  In this respect, the user's attention is drawn to the risks associated
  with loading,  using,  modifying and/or developing or reproducing the
  software by the user in light of its specific status of free software,
  that may mean  that it is complicated to manipulate,  and  that  also
  therefore means  that it is reserved for developers  and  experienced
  professionals having in-depth computer knowledge. Users are therefore
  encouraged to load and test the software's suitability as regards their
  requirements in conditions enabling the security of their systems and/or
  data to be ensured and,  more generally, to use and operate it in the
  same conditions as regards security.
  The fact that you are presently reading this means that you have had
  knowledge of the CeCILL-B license and that you accept its terms.
"""

from tensorflow.python.ops import array_ops
from keras.engine.topology import Layer
import keras.backend as K
import numpy as np

class activation_SegSRGAN(Layer):
    def __init__(self, int_channel=0 , seg_channel=1, activation='sigmoid',is_residual=True, **kwargs):
        self.seg_channel = seg_channel
        self.int_channel = int_channel
        self.activation = activation
        self.is_residual = is_residual
        super(activation_SegSRGAN, self).__init__(**kwargs)

    def build(self, input_shapes):
        super(activation_SegSRGAN, self).build(input_shapes)

    def call(self, inputs):
        recent_input = inputs[0] #pred
        first_input = inputs[1] #im
        
        if self.activation == 'sigmoid':
            segmentation = K.sigmoid(recent_input[:, self.seg_channel, :, :, :])
        else:
            assert('Do not support')
        intensity = recent_input[:, self.int_channel, :, :, :]
        
        # Adding channel
        segmentation = K.expand_dims(segmentation,axis=1)
        intensity = K.expand_dims(intensity,axis=1)
        
        if self.is_residual :
            # residual
            residual_intensity = first_input - intensity # correction de l'image interpolee 
            print("A residual model has been initialized")
        else : 
            residual_intensity = intensity
            print("A non residual model has been initialized")
        
        return  K.concatenate([residual_intensity,segmentation], axis=1)
    
    def compute_output_shape(self, input_shapes):
        return input_shapes[0]

def charbonnier_loss(y_true, y_pred):
    """
    https://en.wikipedia.org/wiki/Huber_loss
    """
    epsilon = 1e-3
    diff = y_true - y_pred
    loss = K.mean(K.sqrt(K.square(diff)+epsilon*epsilon), axis=-1)
    return K.mean(loss)

class InstanceNormalization3D(Layer):
    ''' Thanks for github.com/jayanthkoushik/neural-style 
    and https://github.com/PiscesDream/CycleGAN-keras/blob/master/CycleGAN/layers/normalization.py'''
    def __init__(self, **kwargs):
        super(InstanceNormalization3D, self).__init__(**kwargs)

    def build(self, input_shape):
        self.scale = self.add_weight(name='scale', shape=(input_shape[1],), initializer="one", trainable=True)
        self.shift = self.add_weight(name='shift', shape=(input_shape[1],), initializer="zero", trainable=True)
        super(InstanceNormalization3D, self).build(input_shape)

    def call(self, x):
        def image_expand(tensor):
            return K.expand_dims(K.expand_dims(K.expand_dims(tensor, -1), -1), -1)

        def batch_image_expand(tensor):
            return image_expand(K.expand_dims(tensor, 0))

        hwk = K.cast(x.shape[2] * x.shape[3] * x.shape[4], K.floatx())
        mu = K.sum(x, [-1, -2, -3]) / hwk
        mu_vec = image_expand(mu) 
        sig2 = K.sum(K.square(x - mu_vec), [-1, -2, -3]) / hwk
        y = (x - mu_vec) / (K.sqrt(image_expand(sig2)) + K.epsilon())

        scale = batch_image_expand(self.scale)
        shift = batch_image_expand(self.shift)
        return scale*y + shift 

    def compute_output_shape(self, input_shape):
        return input_shape

def wasserstein_loss(y_true, y_pred):
    """Calculates the Wasserstein loss for a sample batch.
    The Wasserstein loss function is very simple to calculate. In a standard GAN, the discriminator
    has a sigmoid output, representing the probability that samples are real or generated. In Wasserstein
    GANs, however, the output is linear with no activation function! Instead of being constrained to [0, 1],
    the discriminator wants to make the distance between its output for real and generated samples as large as possible.
    The most natural way to achieve this is to label generated samples -1 and real samples 1, instead of the
    0 and 1 used in normal GANs, so that multiplying the outputs by the labels will give you the loss immediately.
    Note that the nature of this loss means that it can be (and frequently will be) less than 0."""
    return K.mean(y_true * y_pred)


def gradient_penalty_loss(y_true, y_pred, averaged_samples, gradient_penalty_weight):
    """Calculates the gradient penalty loss for a batch of "averaged" samples.
    In Improved WGANs, the 1-Lipschitz constraint is enforced by adding a term to the loss function
    that penalizes the network if the gradient norm moves away from 1. However, it is impossible to evaluate
    this function at all points in the input space. The compromise used in the paper is to choose random points
    on the lines between real and generated samples, and check the gradients at these points. Note that it is the
    gradient w.r.t. the input averaged samples, not the weights of the discriminator, that we're penalizing!
    In order to evaluate the gradients, we must first run samples through the generator and evaluate the loss.
    Then we get the gradients of the discriminator w.r.t. the input averaged samples.
    The l2 norm and penalty can then be calculated for this gradient.
    Note that this loss function requires the original averaged samples as input, but Keras only supports passing
    y_true and y_pred to loss functions. To get around this, we make a partial() of the function with the
    averaged_samples argument, and use that for model training."""
    # first get the gradients:
    #   assuming: - that y_pred has dimensions (batch_size, 1)
    #             - averaged_samples has dimensions (batch_size, nbr_features)
    # gradients afterwards has dimension (batch_size, nbr_features), basically
    # a list of nbr_features-dimensional gradient vectors
    gradients = K.gradients(y_pred, averaged_samples)[0]
    # compute the euclidean norm by squaring ...
    gradients_sqr = K.square(gradients)
    #   ... summing over the rows ...
    gradients_sqr_sum = K.sum(gradients_sqr,
                              axis=np.arange(1, len(gradients_sqr.shape)))
    #   ... and sqrt
    gradient_l2_norm = K.sqrt(gradients_sqr_sum)
    # compute lambda * (1 - ||grad||)^2 still for each single sample
    gradient_penalty = gradient_penalty_weight * K.square(1 - gradient_l2_norm)
    # return the mean as loss over all the batch samples
    return K.mean(gradient_penalty)

class ReflectPadding3D(Layer):
    def __init__(self, padding=1, **kwargs):
        super(ReflectPadding3D, self).__init__(**kwargs)
        self.padding = ((padding, padding), (padding, padding), (padding, padding))

    def compute_output_shape(self, input_shape):
        if input_shape[2] is not None:
            dim1 = input_shape[2] + self.padding[0][0] + self.padding[0][1]
        else:
            dim1 = None
        if input_shape[3] is not None:
            dim2 = input_shape[3] + self.padding[1][0] + self.padding[1][1]
        else:
            dim2 = None
        if input_shape[4] is not None:
            dim3 = input_shape[4] + self.padding[2][0] + self.padding[2][1]
        else:
            dim3 = None
        return (input_shape[0],
                input_shape[1],
                dim1,
                dim2,
                dim3)

    def call(self, inputs):
        pattern = [[0, 0], [0, 0], 
                   [self.padding[0][0], self.padding[0][1]],
                   [self.padding[1][0], self.padding[1][1]], 
                   [self.padding[2][0], self.padding[2][1]]]
            
        return array_ops.pad(inputs, pattern, mode= "REFLECT")

    def get_config(self):
        config = {'padding': self.padding}
        base_config = super(ReflectPadding3D, self).get_config()
        return dict(list(base_config.items()) + list(config.items()))
    
    
